@@ -6,7 +6,6 @@ module Data.GI.CodeGen.CtoHaskellMap
   ) where
 
 import qualified Data.Map as M
-import Data.Maybe (catMaybes)
 #if !MIN_VERSION_base(4,11,0)
 import Data.Monoid ((<>))
 #endif
@@ -18,11 +17,12 @@ import Data.GI.CodeGen.API (API(..), Name(..), Callback(..),
                             Enumeration(..), EnumerationMember(..),
                             Interface(..), Object(..),
                             Function(..), Method(..), Struct(..), Union(..),
-                            Signal(..))
+                            Signal(..), Property(..))
 import Data.GI.CodeGen.ModulePath (dotModulePath)
 import Data.GI.CodeGen.SymbolNaming (moduleLocation, lowerName, upperName,
-                                     signalHaskellName, haddockSignalAnchor)
-import Data.GI.CodeGen.Util (ucFirst)
+                                     signalHaskellName, haddockSignalAnchor,
+                                     haddockAttrAnchor, hyphensToCamelCase)
+import Data.GI.CodeGen.Util (ucFirst, lcFirst)
 
 -- | Link to an identifier, module, etc.
 data Hyperlink = ValueIdentifier Text
@@ -56,12 +56,12 @@ cToHaskellMap apis = M.union (M.fromList builtins)
         extractRefs (n, APIObject o) = objectRefs n o
 
         builtins :: [(CRef, Hyperlink)]
-        builtins = [(TypeRef "gboolean", TypeIdentifier "P.Bool"),
+        builtins = [(CTypeRef "gboolean", TypeIdentifier "P.Bool"),
                     (ConstantRef "TRUE", ValueIdentifier "P.True"),
                     (ConstantRef "FALSE", ValueIdentifier "P.False"),
-                    (TypeRef "GError", TypeIdentifier "GError"),
-                    (TypeRef "GType", TypeIdentifier "GType"),
-                    (TypeRef "GVariant", TypeIdentifier "GVariant"),
+                    (CTypeRef "GError", TypeIdentifier "GError"),
+                    (CTypeRef "GType", TypeIdentifier "GType"),
+                    (CTypeRef "GVariant", TypeIdentifier "GVariant"),
                     (ConstantRef "NULL", ValueIdentifier "P.Nothing")]
 
 -- | Obtain the fully qualified symbol pointing to a value.
@@ -77,41 +77,44 @@ fullyQualifiedType n api symbol =
 -- | Extract the C name of a constant. These are often referred to as
 -- types, so we allow that too.
 constRefs :: Name -> Constant -> [(CRef, Hyperlink)]
-constRefs n c = [(ConstantRef (constantCType c),
-                  fullyQualifiedValue n (APIConst c) $ name n),
-                 (TypeRef (constantCType c),
-                  fullyQualifiedValue n (APIConst c) $ name n)]
+constRefs n c = [(ConstantRef (constantCType c), qualified),
+                 (CTypeRef (constantCType c), qualified),
+                 (TypeRef n, qualified)]
+  where qualified = fullyQualifiedValue n (APIConst c) $ name n
 
 -- | Extract the C name of a function.
 funcRefs :: Name -> Function -> [(CRef, Hyperlink)]
-funcRefs n f = [(FunctionRef (fnSymbol f),
-                 fullyQualifiedValue n (APIFunction f) $ lowerName n)]
+funcRefs n f = [(OldFunctionRef (fnSymbol f), qualified),
+                (FunctionRef n, qualified)]
+  where qualified = fullyQualifiedValue n (APIFunction f) $ lowerName n
 
 -- | Extract the C names of the fields in an enumeration/flags, and
 -- the name of the type itself.
 enumRefs :: API -> Name -> Enumeration -> [(CRef, Hyperlink)]
-enumRefs api n e = (TypeRef (enumCType e),
-                    fullyQualifiedType n api $ upperName n) :
-                   map memberToRef (enumMembers e)
-  where memberToRef :: EnumerationMember -> (CRef, Hyperlink)
+enumRefs api n e = (CTypeRef (enumCType e), qualified)
+                   : (TypeRef n, qualified)
+                   : map memberToRef (enumMembers e)
+  where qualified = fullyQualifiedType n api $ upperName n
+        memberToRef :: EnumerationMember -> (CRef, Hyperlink)
         memberToRef em = (ConstantRef (enumMemberCId em),
                           fullyQualifiedValue n api $ upperName $
                           n {name = name n <> "_" <> enumMemberName em})
 
 -- | Refs to the methods for a given owner.
 methodRefs :: Name -> API -> [Method] -> [(CRef, Hyperlink)]
-methodRefs n api methods = catMaybes $ map methodRef methods
-  where methodRef :: Method -> Maybe (CRef, Hyperlink)
+methodRefs n api methods = concatMap methodRef methods
+  where methodRef :: Method -> [(CRef, Hyperlink)]
         methodRef Method{methodSymbol = symbol, methodName = mn} =
           -- Method name namespaced by the owner.
           let mn' = mn {name = name n <> "_" <> name mn}
-          in Just (FunctionRef symbol,
-                   fullyQualifiedValue n api $ lowerName mn')
+              qualified = fullyQualifiedValue n api $ lowerName mn'
+          in [(OldFunctionRef symbol, qualified),
+              (MethodRef n (name mn), qualified)]
 
 -- | Refs to the signals for a given owner.
 signalRefs :: Name -> API -> Maybe Text -> [Signal] -> [(CRef, Hyperlink)]
-signalRefs n api maybeCName signals = map signalRef signals
-  where signalRef :: Signal -> (CRef, Hyperlink)
+signalRefs n@(Name _ owner) api maybeCName signals = concatMap signalRef signals
+  where signalRef :: Signal -> [(CRef, Hyperlink)]
         signalRef (Signal {sigName = sn}) =
           let mod = dotModulePath (moduleLocation n api)
               sn' = signalHaskellName sn
@@ -119,15 +122,34 @@ signalRefs n api maybeCName signals = map signalRef signals
                 Just cname -> cname
                 Nothing -> let Name ns owner = n
                            in ucFirst ns <> owner
-          in (SignalRef ownerCName sn,
-              ModuleLinkWithAnchor (Just sn') mod (haddockSignalAnchor <> sn'))
+              label = Just (owner <> "::" <> sn')
+              link = ModuleLinkWithAnchor label mod (haddockSignalAnchor <> sn')
+          in [(OldSignalRef ownerCName sn, link),
+              (SignalRef n sn, link)]
+
+-- | Refs to the properties for a given owner.
+propRefs :: Name -> API -> Maybe Text -> [Property] -> [(CRef, Hyperlink)]
+propRefs n@(Name _ owner) api maybeCName props = concatMap propertyRef props
+  where propertyRef :: Property -> [(CRef, Hyperlink)]
+        propertyRef (Property {propName = pn}) =
+          let mod = dotModulePath (moduleLocation n api)
+              hn = lcFirst . hyphensToCamelCase $ pn
+              ownerCName = case maybeCName of
+                Just cname -> cname
+                Nothing -> let Name ns owner = n
+                           in ucFirst ns <> owner
+              label = Just (owner <> ":" <> hn)
+              link = ModuleLinkWithAnchor label mod (haddockAttrAnchor <> hn)
+          in [(OldPropertyRef ownerCName pn, link),
+              (PropertyRef n pn, link)]
 
 -- | Given an optional C type and the API constructor construct the
 -- list of associated refs.
 maybeCType :: Name -> API -> Maybe Text -> [(CRef, Hyperlink)]
 maybeCType _ _ Nothing = []
-maybeCType n api (Just ctype) = [(TypeRef ctype,
-                                  fullyQualifiedType n api (upperName n))]
+maybeCType n api (Just ctype) = [(CTypeRef ctype, qualified),
+                                 (TypeRef n, qualified)]
+  where qualified = fullyQualifiedType n api (upperName n)
 
 -- | Extract the C name of a callback.
 callbackRefs :: Name -> Callback -> [(CRef, Hyperlink)]
@@ -154,3 +176,4 @@ objectRefs :: Name -> Object -> [(CRef, Hyperlink)]
 objectRefs n o = maybeCType n (APIObject o) (objCType o)
                  <> methodRefs n (APIObject o) (objMethods o)
                  <> signalRefs n (APIObject o) (objCType o) (objSignals o)
+                 <> propRefs n (APIObject o) (objCType o) (objProperties o)
